@@ -1,7 +1,6 @@
 # -*- coding: utf-8 -*-
-#!/usr/bin/env python
 #
-# Copyright 2012, 2013 BigML
+# Copyright 2014-2016 BigML
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
@@ -26,45 +25,49 @@ import bigmler.resources as r
 import bigmler.checkpoint as c
 
 
-MISSING_TOKENS = ['', 'N/A', 'n/a', 'NULL', 'null', '-', '#DIV/0', '#REF!',
-                  '#NAME?', 'NIL', 'nil', 'NA', 'na', '#VALUE!', '#NULL!',
-                  'NaN', '#N/A', '#NUM!', '?']
 MONTECARLO_FACTOR = 200
 
 
-def ensemble_processing(datasets, objective_field, fields, api, args, resume,
-                        name=None, description=None, model_fields=None,
+def ensemble_processing(datasets, api, args, resume,
+                        fields=None,
                         session_file=None,
                         path=None, log=None):
     """Creates an ensemble of models from the input data
 
     """
     ensembles = []
-    number_of_ensembles = 1
-    if resume:
-        message = u.dated("Ensemble not found. Resuming.\n")
-        resume, ensembles = c.checkpoint(
-            c.are_ensembles_created, path, number_of_ensembles,
-            debug=args.debug,
-            message=message, log_file=session_file, console=args.verbosity)
-    try:
-        ensemble = ensembles[0]
-    except IndexError:
-        ensemble = None
+    ensemble_ids = []
+    models = []
+    model_ids = []
+    number_of_ensembles = len(datasets)
 
-    if ensemble is None:
-        ensemble_args = r.set_ensemble_args(name, description, args,
-                                            model_fields, objective_field,
-                                            fields)
+    if resume:
+        resume, ensemble_ids = c.checkpoint(
+            c.are_ensembles_created, path, number_of_ensembles,
+            debug=args.debug)
+        _, model_ids = c.checkpoint(c.are_models_created, path,
+            number_of_ensembles * args.number_of_models)
+        models = model_ids
+        if not resume:
+            message = u.dated("Found %s ensembles out of %s. Resuming.\n"
+                              % (len(ensemble_ids),
+                                 number_of_ensembles))
+            u.log_message(message, log_file=session_file,
+                          console=args.verbosity)
+        ensembles = ensemble_ids
+        number_of_ensembles -= len(ensemble_ids)
+
+    if number_of_ensembles > 0:
+        ensemble_args = r.set_ensemble_args(args, fields=fields)
         ensembles, ensemble_ids, models, model_ids = r.create_ensembles(
             datasets, ensembles, ensemble_args, args, api=api, path=path,
+            number_of_ensembles=number_of_ensembles,
             session_file=session_file, log=log)
     return ensembles, ensemble_ids, models, model_ids, resume
 
 
-def ensemble_per_label(labels, all_labels, dataset, fields,
-                       objective_field, api, args, resume, name=None,
-                       description=None, model_fields=None,
+def ensemble_per_label(labels, dataset, api, args, resume, fields=None,
+                       multi_label_data=None,
                        session_file=None, path=None, log=None):
     """Creates an ensemble per label for multi-label datasets
 
@@ -89,12 +92,12 @@ def ensemble_per_label(labels, all_labels, dataset, fields,
                           console=args.verbosity)
             # erase models' info that will be rebuilt
             u.log_created_resources("models", path, None,
-                                    open_mode='w')
+                                    mode='w')
     number_of_ensembles = len(labels) - len(ensemble_ids)
     ensemble_args_list = r.set_label_ensemble_args(
-        name, description, args,
-        labels, all_labels, number_of_ensembles,
-        fields, model_fields, objective_field)
+        args,
+        labels, multi_label_data, number_of_ensembles,
+        fields)
 
     # create ensembles changing the input_field to select
     # only one label at a time
