@@ -1,7 +1,6 @@
 # -*- coding: utf-8 -*-
-#!/usr/bin/env python
 #
-# Copyright 2012, 2013 BigML
+# Copyright 2014-2016 BigML
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
@@ -20,61 +19,23 @@
 """
 from __future__ import absolute_import
 
-import sys
 import copy
 
 import bigml.api
 import bigmler.utils as u
 import bigmler.resources as r
 import bigmler.checkpoint as c
-import bigmler.labels as l
 
-from bigml.fields import Fields
+from bigml.fields import Fields, DEFAULT_MISSING_TOKENS
 
 from bigmler.processing.ensembles import (ensemble_processing,
                                           ensemble_per_label)
 
-from bigmler.processing.ensembles import MISSING_TOKENS
-
 MONTECARLO_FACTOR = 200
 
 
-def has_models(args):
-    """Returns if some kind of model or ensemble is given in args.
-
-    """
-    return (args.model or args.ensemble or args.ensembles
-            or args.models or args.model_tag or args.ensemble_tag)
-
-
-def set_multi_label_objective(fields_dict, objective):
-    """Changes the field information for the objective field
-       in the fields_dict dictionnary to the real objective attributes for
-       multi-label models.
-
-    """
-    target_field = fields_dict[objective]
-    if target_field['label'].startswith(l.MULTI_LABEL_LABEL):
-        label = target_field['label'][len(l.MULTI_LABEL_LABEL):]
-        objective_name = target_field['name']
-        suffix_length = len(label) + 3
-        try:
-            objective_name = objective_name[0: -suffix_length]
-            target_field['name'] = objective_name
-            target_field['label'] = 'multi-label objective'
-        except IndexError:
-            sys.exit("It seems that the label of multi-labeled fields has"
-                     " been altered. You should not change the labels of"
-                     " generated fields.")
-    else:
-        sys.exit("It seems that the label of multi-labeled fields has been"
-                 " altered. You should not change the labels of generated"
-                 " fields.")
-
-
-def model_per_label(labels, all_labels, datasets, fields,
-                    objective_field, api, args, resume, name=None,
-                    description=None, model_fields=None,
+def model_per_label(labels, datasets, api, args, resume, fields=None,
+                    multi_label_data=None,
                     session_file=None, path=None, log=None):
     """Creates a model per label for multi-label datasets
 
@@ -96,9 +57,8 @@ def model_per_label(labels, all_labels, datasets, fields,
 
         models = model_ids
     args.number_of_models = len(labels) - len(model_ids)
-    model_args_list = r.set_label_model_args(
-        name, description, args,
-        labels, all_labels, fields, model_fields, objective_field)
+    model_args_list = r.set_label_model_args(args, fields, labels,
+                                             multi_label_data)
 
     # create models changing the input_field to select
     # only one label at a time
@@ -109,11 +69,11 @@ def model_per_label(labels, all_labels, datasets, fields,
     return models, model_ids, resume
 
 
-def models_processing(datasets, models, model_ids, objective_field, fields,
-                      api, args, resume,
-                      name=None, description=None, model_fields=None,
+def models_processing(datasets, models, model_ids,
+                      api, args, resume, fields=None,
                       session_file=None, path=None,
-                      log=None, labels=None, all_labels=None):
+                      log=None, labels=None, multi_label_data=None,
+                      other_label=None):
     """Creates or retrieves models from the input data
 
     """
@@ -121,41 +81,33 @@ def models_processing(datasets, models, model_ids, objective_field, fields,
 
     # If we have a dataset but not a model, we create the model if the no_model
     # flag hasn't been set up.
-    if datasets and not (has_models(args) or args.no_model):
+    if datasets and not (args.has_models_ or args.no_model):
         dataset = datasets[0]
         model_ids = []
         models = []
         if args.multi_label:
-            # Create one model per column choosing only the label column
-            if args.training_set is None:
-                if objective_field is None:
-                    objective_field = fields.objective_field
-                objective_id = fields.field_id(objective_field)
-                objective_name = fields.field_name(objective_id)
-                all_labels, labels = l.retrieve_labels(fields.fields, labels,
-                                                       objective_name)
             # If --number-of-models is not set or is 1, create one model per
             # label. Otherwise, create one ensemble per label with the required
             # number of models
             if args.number_of_models < 2:
                 models, model_ids, resume = model_per_label(
-                    labels, all_labels, datasets, fields,
-                    objective_field, api, args, resume, name, description,
-                    model_fields, session_file, path, log)
+                    labels, datasets, api, args, resume,
+                    fields=fields, multi_label_data=multi_label_data,
+                    session_file=session_file, path=path, log=log)
             else:
                 (ensembles, ensemble_ids,
                  models, model_ids, resume) = ensemble_per_label(
-                     labels, all_labels, dataset, fields,
-                     objective_field, api, args, resume, name, description,
-                     model_fields, session_file, path, log)
+                     labels, dataset, api, args, resume, fields=fields,
+                     multi_label_data=multi_label_data,
+                     session_file=session_file, path=path, log=log)
 
         elif args.number_of_models > 1:
             ensembles = []
-            # Ensemble of models
+            # Ensembles of models
             (ensembles, ensemble_ids,
              models, model_ids, resume) = ensemble_processing(
-                 datasets, objective_field, fields, api, args, resume,
-                 name=name, description=description, model_fields=model_fields,
+                 datasets, api, args, resume,
+                 fields=fields,
                  session_file=session_file, path=path, log=log)
             ensemble = ensembles[0]
             args.ensemble = bigml.api.get_ensemble_id(ensemble)
@@ -164,6 +116,9 @@ def models_processing(datasets, models, model_ids, objective_field, fields,
             # Set of partial datasets created setting args.max_categories
             if len(datasets) > 1 and args.max_categories:
                 args.number_of_models = len(datasets)
+            if ((args.test_datasets and args.evaluate) or
+                    (args.datasets and args.evaluate and args.dataset_off)):
+                args.number_of_models = len(args.dataset_ids)
             # Cross-validation case: we create 2 * n models to be validated
             # holding out an n% of data
             if args.cross_validation_rate > 0:
@@ -185,10 +140,11 @@ def models_processing(datasets, models, model_ids, objective_field, fields,
 
                 models = model_ids
                 args.number_of_models -= len(model_ids)
-
-            model_args = r.set_model_args(name, description, args,
-                                          objective_field, fields,
-                                          model_fields)
+            model_args = r.set_model_args(args,
+                                          fields=fields,
+                                          objective_id=args.objective_id_,
+                                          model_fields=args.model_fields_,
+                                          other_label=other_label)
             models, model_ids = r.create_models(datasets, models,
                                                 model_args, args, api,
                                                 path, session_file, log)
@@ -201,12 +157,13 @@ def models_processing(datasets, models, model_ids, objective_field, fields,
         models = model_ids[:]
 
     if args.ensemble:
-        ensemble = r.get_ensemble(args.ensemble, api, args.verbosity,
-                                  session_file)
-        ensemble_ids = [ensemble]
-        model_ids = ensemble['object']['models']
-
-        models = model_ids[:]
+        if not args.ensemble in ensemble_ids:
+            ensemble_ids.append(args.ensemble)
+        if not args.evaluate:
+            ensemble = r.get_ensemble(args.ensemble, api, args.verbosity,
+                                      session_file)
+            model_ids = ensemble['object']['models']
+            models = model_ids[:]
 
     if args.ensembles or args.ensemble_tag:
         model_ids = []
@@ -232,7 +189,8 @@ def models_processing(datasets, models, model_ids, objective_field, fields,
     return models, model_ids, ensemble_ids, resume
 
 
-def get_model_fields(model, csv_properties, args, single_model=True):
+def get_model_fields(model, csv_properties, args, single_model=True,
+                     multi_label_data=None):
     """Retrieves fields info from model resource
 
     """
@@ -242,7 +200,8 @@ def get_model_fields(model, csv_properties, args, single_model=True):
     if args.user_locale is None:
         args.user_locale = model['object'].get('locale', None)
     csv_properties.update(data_locale=args.user_locale)
-    if single_model and 'model_fields' in model['object']['model']:
+    if (single_model and 'model_fields' in model['object']['model'] and
+            args.test_header):
         model_fields = model['object']['model']['model_fields'].keys()
         csv_properties.update(include=model_fields)
     else:
@@ -250,21 +209,25 @@ def get_model_fields(model, csv_properties, args, single_model=True):
     if 'missing_tokens' in model['object']['model']:
         missing_tokens = model['object']['model']['missing_tokens']
     else:
-        missing_tokens = MISSING_TOKENS
+        missing_tokens = DEFAULT_MISSING_TOKENS
     csv_properties.update(missing_tokens=missing_tokens)
     # if the model belongs to a multi-label set of models, the real objective
     # field is never amongst the set of fields of each individual model, so
     # we must add it.
     fields_dict = copy.deepcopy(model['object']['model']['fields'])
-    objective_field = model['object']['objective_fields']
-    if isinstance(objective_field, list):
-        objective_field = objective_field[0]
+
     if args.multi_label:
-        # Changes fields_dict objective field attributes to the real
-        # multi-label objective
-        set_multi_label_objective(fields_dict, objective_field)
+        # Adds the real objective field to fields_dict
+        objective_field = multi_label_data['objective_name']
+        objective_id = multi_label_data['objective_id']
+        objective_column = multi_label_data['objective_column']
+        fields_dict[objective_id] = {"op_type": "categorical",
+                                     "name": objective_field,
+                                     "column_number": objective_column}
+    else:
+        objective_field = model['object']['objective_fields']
+        if isinstance(objective_field, list):
+            objective_field = objective_field[0]
     csv_properties.update(objective_field=objective_field)
-
-    fields = Fields(fields_dict, **csv_properties)
-
-    return fields, objective_field
+    args.objective_field = objective_field
+    return Fields(fields_dict, **csv_properties)
