@@ -1,7 +1,6 @@
 # -*- coding: utf-8 -*-
-#!/usr/bin/env python
 #
-# Copyright 2013 BigML
+# Copyright 2013-2016 BigML
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
@@ -22,12 +21,23 @@
 """
 from __future__ import absolute_import
 
-import csv
 import sys
 
 from bigml.util import get_csv_delimiter
+from bigml.io import UnicodeReader
 
 from bigmler.checkpoint import file_number_of_lines
+from bigmler.labels import get_label_field
+from bigmler.utils import PYTHON3, SYSTEM_ENCODING, FILE_ENCODING
+from bigmler.utils import encode2, decode2
+from bigmler.utf8recoder import UTF8Recoder
+
+
+AGGREGATES = {
+    'count': len,
+    'last': lambda x: x[-1],
+    'first': lambda x: x[0]
+}
 
 
 class TrainReader(object):
@@ -36,7 +46,8 @@ class TrainReader(object):
     """
     def __init__(self, training_set, training_set_header, objective_field,
                  multi_label=False, labels=None, label_separator=None,
-                 training_separator=None, multi_label_fields=None):
+                 training_separator=None, multi_label_fields=None,
+                 label_aggregates=None, objective=True):
         """Builds a generator from a csv file
 
            `training_set`: path to the training data file
@@ -46,39 +57,76 @@ class TrainReader(object):
            `labels`: Fields object with the expected fields structure.
         """
         self.training_set = training_set
+
+        if training_set.__class__.__name__ == "StringIO":
+            self.encode = None
+            self.training_set = UTF8Recoder(training_set, SYSTEM_ENCODING)
+        else:
+            self.encode = None if PYTHON3 else FILE_ENCODING
         self.training_set_header = training_set_header
-        self.training_set_handler = None
         self.training_reader = None
         self.multi_label = multi_label
-
-        self.training_separator = (training_separator.decode("string_escape")
+        self.objective = objective
+        if label_aggregates is None:
+            label_aggregates = []
+        self.label_aggregates = label_aggregates
+        self.training_separator = (decode2(training_separator,
+                                           encoding="string_escape")
                                    if training_separator is not None
                                    else get_csv_delimiter())
         if len(self.training_separator) > 1:
             sys.exit("Only one character can be used as test data separator.")
         # opening csv reader
         self.reset()
-        self.label_separator = (label_separator.decode("string_escape")
+        self.label_separator = (decode2(label_separator,
+                                        encoding="string_escape")
                                 if label_separator is not None
                                 else get_csv_delimiter())
 
-        first_row = self.next(reset=not training_set_header)
+        first_row = self.get_next(reset=not training_set_header)
         self.row_length = len(first_row)
 
         if training_set_header:
-            self.headers = [unicode(header, "utf-8") for header in first_row]
+            self.headers = first_row
         else:
             self.headers = [("field_%s" % index) for index in
                             range(0, self.row_length)]
 
         self.multi_label_fields = sorted(self._get_columns(multi_label_fields))
-        self.objective_column = self._get_columns([objective_field])[0]
-        if not self.objective_column in self.multi_label_fields:
-            self.multi_label_fields.append(self.objective_column)
-        self.labels = labels
+        if objective:
+            self.objective_column = self._get_columns([objective_field])[0]
+            if not self.objective_column in self.multi_label_fields:
+                self.multi_label_fields.append(self.objective_column)
+            self.labels = labels
         self.fields_labels = self._get_labels()
-        self.labels = self.fields_labels[self.objective_column]
-        self.objective_name = self.headers[self.objective_column]
+        if objective:
+            if labels is None:
+                self.labels = self.fields_labels[self.objective_column]
+            self.objective_name = self.headers[self.objective_column]
+
+    def __iter__(self):
+        """Iterator method
+
+        """
+        return self
+
+    def get_label_headers(self):
+        """Returns a list of headers with the new extended field names for
+           each objective label
+        """
+        new_headers = self.get_headers()
+        for field_column in self.multi_label_fields:
+            labels = self.fields_labels[field_column]
+            new_field_names = [get_label_field(self.headers[field_column],
+                                               label)
+                               for label in labels]
+            new_headers.extend(new_field_names)
+            for aggregate in self.label_aggregates:
+                new_headers.append(get_label_field(
+                    self.headers[field_column], aggregate))
+        if not PYTHON3:
+            new_headers = [encode2(header) for header in new_headers]
+        return new_headers
 
     def _get_columns(self, fields_list):
         """Receives a comma-separated list of fields given by name or
@@ -100,12 +148,17 @@ class TrainReader(object):
                 try:
                     column = self.headers.index(field)
                 except ValueError:
-                    sys.exit("The %s has been set as multi-label field but"
-                             " it cannot be found in the headers row: \n %s" %
-                             (field,
-                              ", ".join([header.encode("utf-8")
-                                         for header in self.headers])))
-            column_list.append(column)
+                    if self.objective:
+                        sys.exit("The %s has been set as multi-label field but"
+                                 " it cannot be found in the headers row: \n"
+                                 " %s" %
+                                 (field,
+                                  ", ".join([encode2(header)
+                                             for header in self.headers])))
+                    else:
+                        column = None
+            if column is not None:
+                column_list.append(column)
         return column_list
 
     def reset(self):
@@ -113,24 +166,23 @@ class TrainReader(object):
 
         """
         try:
-            self.training_set_handler.close()
+            self.training_set.close()
         except (IOError, AttributeError):
             pass
         try:
-            self.training_set_handler = open(self.training_set, "U")
-            self.training_reader = csv.reader(
-                self.training_set_handler, delimiter=self.training_separator,
-                lineterminator="\n")
+            self.training_reader = UnicodeReader(
+                self.training_set, delimiter=self.training_separator,
+                lineterminator="\n").open_reader()
         except IOError:
             sys.exit("Error: cannot read training %s" % self.training_set)
 
-    def __iter__(self):
-        """Iterator method
+    def next(self):
+        """Iterator method for next item
 
         """
-        return self
+        return self.get_next()
 
-    def next(self, extended=False, reset=False):
+    def get_next(self, extended=False, reset=False):
         """Returns the next row. If extended is True, the row is extended with
            a list of booleans depending on whether the label is in the
            objective field value or not. If reset is True, the file is
@@ -140,23 +192,26 @@ class TrainReader(object):
         row = self.training_reader.next()
         row = [value.strip() for value in row]
         if extended:
-            if self.multi_label and self.labels is None:
+            if self.multi_label and self.fields_labels is None:
                 self.fields_labels = self._get_labels()
 
             for field_column in self.multi_label_fields:
                 aggregated_field_value = row[field_column]
                 field_values = aggregated_field_value.split(
                     self.label_separator)
-                field_values = [value.decode("utf-8").strip() for
+
+                field_values = [value.strip() for
                                 value in field_values]
-                labels_row = [label in field_values for label in
+
+                labels_row = [int(label in field_values) for label in
                               self.fields_labels[field_column]]
                 row.extend(labels_row)
-            aggregated_value = row[self.objective_column]
-            del row[self.objective_column]
-            row.append(aggregated_value.strip())
+                for aggregate in self.label_aggregates:
+                    row.append(AGGREGATES[aggregate](field_values))
         if reset:
             self.reset()
+        if not PYTHON3:
+            row = [encode2(item) for item in row]
         return row
 
     def number_of_rows(self):
@@ -179,7 +234,6 @@ class TrainReader(object):
 
         """
         labels = {}
-        labels[self.objective_column] = []
         for field_column in self.multi_label_fields:
             labels[field_column] = []
         for row in self:
@@ -187,8 +241,6 @@ class TrainReader(object):
                 labels = self._get_field_labels(row, labels,
                                                 field_column,
                                                 self.label_separator)
-        if self.labels is None:
-            self.labels = labels[self.objective_column]
         return labels
 
     def _get_field_labels(self, row, labels, field_column, separator):
@@ -198,14 +250,18 @@ class TrainReader(object):
         field_value = row[field_column]
         if self.multi_label:
             new_labels = field_value.split(separator)
-            new_labels = map(str.strip, new_labels)
-            new_labels = [label.decode("utf-8").strip()
+            new_labels = [decode2(label).strip()
                           for label in new_labels]
             # TODO: clean user given missing tokens
             for label_index in range(0, len(new_labels)):
                 if new_labels[label_index] == '':
-                    del(new_labels[label_index])
+                    del new_labels[label_index]
             if new_labels != []:
+                if (self.objective and field_column == self.objective_column
+                        and self.labels is not None):
+                    # If user gave the subset of labels, use only those
+                    new_labels = [label for label in self.labels if
+                                  label in new_labels]
                 labels[field_column].extend(new_labels)
         else:
             labels[field_column].append(field_value)
@@ -220,19 +276,44 @@ class TrainReader(object):
         if objective_field:
             return self.headers[:]
         new_headers = self.headers[:]
-        del new_headers[self.objective_column]
+        if self.objective:
+            del new_headers[self.objective_column]
         return new_headers
 
-    def labels_columns(self):
-        """List of tuples (label_column, label) describing the label extension
+    def new_fields_info(self):
+        """Dict of 2-item lists 'field_column': [label, label_column]
+           describing the per label extension
 
         """
-        columns = []
-        column = len(self.headers) - 1
+        info = {}
+        column = len(self.headers)
         for field_column in self.multi_label_fields:
+            alpha_field_column = str(field_column)
+            info[alpha_field_column] = []
             labels = self.fields_labels[field_column]
             for label in labels:
-                columns.append((column, label))
+                info[alpha_field_column].append([label, column])
                 column += 1
+            # skip the aggregate values columns
+            column += len(self.label_aggregates)
 
-        return columns
+        return info
+
+    def get_multi_label_data(self):
+        """Returns a dict to store the multi-label info that defines this
+           source
+
+        """
+        if self.objective:
+            return {
+                "multi_label_fields": [[column, self.headers[column]]
+                                       for column in self.multi_label_fields],
+                "generated_fields": self.new_fields_info(),
+                "objective_name": self.objective_name,
+                "objective_column": self.objective_column}
+
+    def close(self):
+        """Closing file handler
+
+        """
+        self.training_reader.close_reader()
