@@ -1,7 +1,6 @@
 # -*- coding: utf-8 -*-
-#!/usr/bin/env python
 #
-# Copyright 2012, 2013 BigML
+# Copyright 2012-2016 BigML
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
@@ -20,7 +19,6 @@
 """
 from __future__ import absolute_import
 
-import csv
 import fileinput
 import ast
 import glob
@@ -34,19 +32,32 @@ except ImportError:
     import json
 
 import bigml.api
-from bigml.util import console_log
+from bigml.util import console_log, empty_resource
+from bigml.fields import get_fields_structure, Fields
+from bigml.io import UnicodeReader
 
+PYTHON3 = sys.version_info[0] == 3
 PAGE_LENGTH = 200
 ATTRIBUTE_NAMES = ['name', 'label', 'description']
-NEW_DIRS_LOG = ".bigmler_dirs"
+NEW_DIRS_LOG = u".bigmler_dirs"
 BRIEF_MODEL_QS = "exclude=root,fields"
 
 # Base Domain
 BIGML_DOMAIN = os.environ.get('BIGML_DOMAIN', 'bigml.io')
 BIGML_DASHBOARD_URL = os.environ.get('BIGML_DASHBOARD_URL')
-RESOURCE_URL = ("https://%s/dashboard/" % (BIGML_DOMAIN[:-3] + '.com')
+BIGML_DASHBOARD_DOMAIN = (
+    "%s.com" % BIGML_DOMAIN[:-3] if BIGML_DOMAIN[-3:] == '.io' else
+    BIGML_DOMAIN.replace("-io.", "."))
+RESOURCE_URL = ("https://%s/dashboard/" % BIGML_DASHBOARD_DOMAIN
                 if BIGML_DASHBOARD_URL is None
                 else BIGML_DASHBOARD_URL)
+RESOURCE_SHARED_URL = "%s/shared/" % "/".join(RESOURCE_URL.split('/')[:-2])
+RESOURCE_EMBEDDED_URL = ("%s/embedded/%%s/tree" %
+                         "/".join(RESOURCE_URL.split('/')[:-2]))
+
+
+SYSTEM_ENCODING = 'cp1252' if sys.platform == 'win32' else 'utf-8'
+FILE_ENCODING = 'utf-8'
 
 
 def read_description(path):
@@ -71,19 +82,20 @@ def read_field_attributes(path):
     """
     field_attributes = {}
     try:
-        attributes_reader = csv.reader(open(path, "U"), quotechar="'")
+        with UnicodeReader(path, quotechar="'") as attributes_reader:
+            for row in attributes_reader:
+                attributes = {}
+                if len(row) > 1:
+                    for index in range(0, min(len(ATTRIBUTE_NAMES),
+                                              len(row) - 1)):
+                        attributes.update(
+                            {ATTRIBUTE_NAMES[index]: row[index + 1]})
+                    field_attributes.update({
+                        int(row[0]): attributes})
+            return field_attributes
     except IOError:
         sys.exit("Error: cannot read field attributes %s" % path)
 
-    for row in attributes_reader:
-        attributes = {}
-        if len(row) > 1:
-            for index in range(0, min(len(ATTRIBUTE_NAMES), len(row) - 1)):
-                attributes.update({ATTRIBUTE_NAMES[index]: row[index + 1]})
-            field_attributes.update({
-                int(row[0]): attributes})
-    return field_attributes
-
 
 def read_types(path):
     """Types to update source fields types.
@@ -122,7 +134,7 @@ def read_json(path):
     """
     json_attributes = {}
     try:
-        attributes_reader = open(path, "r")
+        attributes_reader = open(path, open_mode('r'))
     except IOError:
         sys.exit("Error: cannot read json file %s" % path)
     try:
@@ -224,10 +236,9 @@ def read_votes_files(dirs_list, path):
     Models' predictions files are expected to be named after the model id,
     for instance: model_50974922035d0706da00003d__predictions.csv
     """
-    file_name = "%s%scombined_predictions" % (path,
-                                              os.sep)
+    file_name = "%s%scombined_predictions" % (path, os.sep)
     check_dir(file_name)
-    group_predictions = open(file_name, "w", 0)
+    group_predictions = open(file_name, "wb", 0)
     current_directory = os.getcwd()
     predictions_files = []
     for directory in dirs_list:
@@ -235,27 +246,67 @@ def read_votes_files(dirs_list, path):
         os.chdir(directory)
         for predictions_file in glob.glob("model_*_predictions.csv"):
             predictions_files.append("%s%s%s" % (os.getcwd(),
-                                     os.sep, predictions_file))
-            group_predictions.write("%s\n" % predictions_file)
+                                                 os.sep, predictions_file))
+            message = "%s\n" % predictions_file
+            if PYTHON3:
+                message = message.encode(FILE_ENCODING)
+            group_predictions.write(message)
         os.chdir(current_directory)
     group_predictions.close()
     return predictions_files
 
 
-def list_ids(api_function, query_string):
+def read_local_resource(path, csv_properties=None):
+    """Read the JSON resource structure information from the given file.
+
+    """
+    resource = empty_resource()
+    if csv_properties is None:
+        csv_properties = {}
+    fields = None
+    with open(path, open_mode('r')) as resource_file:
+        try:
+            resource = json.loads(resource_file.read())
+        except IOError:
+            pass
+    resource_id = resource.get('resource')
+    if resource_id is None:
+        sys.exit("Failed to extract a BigML resource structure from the"
+                 " contents of file %s." % path)
+    if resource.get('object') is None:
+        resource = {'resource': resource_id,
+                    'object': resource,
+                    'error': None,
+                    'code': bigml.api.HTTP_OK}
+
+    fields, resource_locale, missing_tokens, \
+    objective_column = get_fields_structure(resource)
+    if not 'objective_field' in csv_properties and \
+            objective_column is not None:
+        csv_properties['objective_field'] = objective_column
+    if missing_tokens:
+        csv_properties['missing_tokens'] = missing_tokens
+    if resource_locale:
+        csv_properties['data_locale'] = resource_locale
+    if fields:
+        fields = Fields(resource, **csv_properties)
+    return resource, csv_properties, fields
+
+
+def list_ids(api_function, query_string, status_code=bigml.api.FINISHED):
     """Lists BigML resources filtered by `query_string`.
 
     """
     q_s = 'status.code=%s;limit=%s;%s' % (
-          bigml.api.FINISHED, PAGE_LENGTH, query_string)
+        status_code, PAGE_LENGTH, query_string)
     resources = api_function(q_s)
     ids = [obj['resource'] for obj in (resources['objects'] or [])]
     while (resources['objects'] and
-          (resources['meta']['total_count'] > (resources['meta']['offset'] +
-           resources['meta']['limit']))):
+           (resources['meta']['total_count'] > (resources['meta']['offset'] +
+                                                resources['meta']['limit']))):
         offset = resources['meta']['offset'] + PAGE_LENGTH
         q_s = 'status.code=%s;offset=%s;limit=%s;%s' % (
-              bigml.api.FINISHED, offset, PAGE_LENGTH, query_string)
+            status_code, offset, PAGE_LENGTH, query_string)
         resources = api_function(q_s)
         if resources['objects']:
             ids.extend([obj['resource'] for obj in resources['objects']])
@@ -266,22 +317,16 @@ def delete(api, delete_list):
     """ Deletes the resources given in the list.
 
     """
-    delete_functions = {bigml.api.SOURCE_RE: api.delete_source,
-                        bigml.api.DATASET_RE: api.delete_dataset,
-                        bigml.api.MODEL_RE: api.delete_model,
-                        bigml.api.PREDICTION_RE: api.delete_prediction,
-                        bigml.api.EVALUATION_RE: api.delete_evaluation,
-                        bigml.api.ENSEMBLE_RE: api.delete_ensemble}
     for resource_id in delete_list:
         resource_type = None
         try:
-            for resource_type in delete_functions:
+            for resource_type in bigml.api.RESOURCE_RE.keys():
                 try:
                     bigml.api.get_resource(resource_type, resource_id)
                     break
                 except ValueError:
                     pass
-            delete_functions[resource_type](resource_id)
+            api.deleters[resource_type](resource_id)
         except ValueError:
             console_log("Failed to delete resource %s" % resource_id)
 
@@ -293,59 +338,21 @@ def check_dir(path):
     directory = os.path.dirname(path)
     if len(directory) > 0 and not os.path.exists(directory):
         os.makedirs(directory)
-        with open(NEW_DIRS_LOG, "a", 0) as directory_log:
-            directory_log.write("%s\n" % os.path.abspath(directory))
+        sys_log_message(u"%s\n" % os.path.abspath(directory),
+                        log_file=NEW_DIRS_LOG)
     return directory
 
 
-def tail(file_handler, window=1):
-    """Returns the last n lines of a file.
-
-    """
-    bufsiz = 1024
-    file_handler.seek(0, 2)
-    file_bytes = file_handler.tell()
-    size = window + 1
-    block = -1
-    data = []
-    while size > 0 and file_bytes > 0:
-        if (file_bytes - bufsiz > 0):
-            # Seek back one whole bufsiz
-            file_handler.seek(block * bufsiz, 2)
-            # read BUFFER
-            new_data = [file_handler.read(bufsiz)]
-            new_data.extend(data)
-            data = new_data
-        else:
-            # file too small, start from begining
-            file_handler.seek(0, 0)
-            # only read what was not read
-            data.append(file_handler.read(file_bytes))
-        lines_found = data[0].count('\n')
-        size -= lines_found
-        file_bytes -= bufsiz
-        block -= 1
-    return ''.join(data).splitlines()[-window:]
-
-
-def get_log_reversed(file_name, stack_level):
-    """Reads the line of a log file that has the chosen stack_level
-
-    """
-    lines_list = tail(open(file_name, "r"), window=(stack_level + 1))
-    return lines_list[0]
-
-
 def print_tree(directory, padding):
     """Returns a graphical directory tree structure as a string
 
     """
-    if padding != ' ':
-        output = padding[:-1] + '├─'
+    if padding != u' ':
+        output = padding[:-1] + u'├─'
     else:
         output = padding
-    output += os.path.basename(os.path.abspath(directory)) + '\n'
-    padding = padding + ' '
+    output += os.path.basename(os.path.abspath(directory)) + u'\n'
+    padding = padding + u' '
     files = []
     files = os.listdir(directory)
     count = 0
@@ -355,14 +362,14 @@ def print_tree(directory, padding):
         path = directory + os.sep + file_name
         if os.path.isdir(path):
             if count == len(files):
-                output += print_tree(path, padding + ' ')
+                output += print_tree(path, padding + u' ')
             else:
-                output += print_tree(path, padding + '|')
+                output += print_tree(path, padding + u'|')
         else:
             if i < (len(files) - 1):
-                output += padding + '├─' + file_name + '\n'
+                output += padding + u'├─' + file_name + u'\n'
             else:
-                output += padding + '└─' + file_name + '\n'
+                output += padding + u'└─' + file_name + u'\n'
     return output
 
 
@@ -374,14 +381,22 @@ def dated(message):
                         message)
 
 
-def get_url(resource):
+def get_url(resource, shared=False, embedded=False):
     """Returns the resource's url in bigml.com
 
     """
-    resource_id = bigml.api.get_resource_id(resource)
-    if not resource_id:
-        return ""
-    return RESOURCE_URL + resource_id
+    if shared:
+        return (RESOURCE_SHARED_URL + bigml.api.get_resource_type(resource)
+                + os.sep + resource['object']['shared_hash'])
+    elif embedded:
+        return (RESOURCE_EMBEDDED_URL % (bigml.api.get_resource_type(resource)
+                                         + os.sep +
+                                         resource['object']['shared_hash']))
+    else:
+        resource_id = bigml.api.get_resource_id(resource)
+        if not resource_id:
+            return ""
+        return RESOURCE_URL + resource_id
 
 
 def log_message(message, log_file=None, console=False):
@@ -390,12 +405,27 @@ def log_message(message, log_file=None, console=False):
        If log_file is set, logs the message in the file.
        If console is True, sends the message to console.
     """
+
+    if isinstance(message, unicode) and not PYTHON3:
+        message = message.encode(FILE_ENCODING)
     if console:
         console_log(message)
     if log_file is not None:
-        if isinstance(message, unicode):
-            message = message.encode("utf-8")
-        with open(log_file, 'a', 0) as log_file:
+        if PYTHON3:
+            message = message.encode(FILE_ENCODING)
+        with open(log_file, 'ab', 0) as log_file:
+            log_file.write(message)
+
+
+def sys_log_message(message, log_file=None):
+    """Logs a message in a file using the system encoding
+
+       If log_file is set, logs the message in the file.
+    """
+    if isinstance(message, unicode):
+        message = message.encode(SYSTEM_ENCODING)
+    if log_file is not None:
+        with open(log_file, 'ab', 0) as log_file:
             log_file.write(message)
 
 
@@ -420,18 +450,26 @@ def check_resource_error(resource, message):
     return bigml.api.get_resource_id(resource)
 
 
-def log_created_resources(file_name, path, resource_id, open_mode='w'):
+def log_created_resources(file_name, path, resource_id, mode='w',
+                          comment=None):
     """Logs the created resources ids in the given file
 
     """
     if path is not None:
         file_name = "%s%s%s" % (path, os.sep, file_name)
         try:
-            with open(file_name, open_mode, 0) as resource_file:
+            with open(file_name, "%sb" % mode, 0) as resource_file:
                 if resource_id is not None:
-                    resource_file.write("%s\n" % resource_id)
+                    message = "%s\n" % resource_id
+                    if PYTHON3 or isinstance(message, unicode):
+                        message = message.encode(SYSTEM_ENCODING)
+                    resource_file.write(message)
+                if comment is not None:
+                    if PYTHON3 or isinstance(comment, unicode):
+                        comment = comment.encode(SYSTEM_ENCODING)
+                    resource_file.write(comment)
         except IOError, exc:
-            print ("Failed to write %s: %s" % (file_name, str(exc)))
+            print "Failed to write %s: %s" % (file_name, str(exc))
 
 
 def check_resource(*args, **kwargs):
@@ -512,3 +550,160 @@ def objective_field_name(model_or_ensemble, api):
     else:
         sys.exit("No valid model or ensemble structure.")
     return model['object']['model']['model_fields'][objective_field]['name']
+
+
+def read_objective_weights(path):
+    """Reads objective weights from a CSV file in a class, weight format.
+
+    The expected structure is:
+    class name, weight
+
+    For example:
+
+    Iris-setosa,5
+    Iris-versicolor,10
+
+    """
+    objective_weights = []
+    try:
+        with UnicodeReader(path, quotechar="'") as weights_reader:
+            for row in weights_reader:
+                weights = []
+                if len(row) != 2:
+                    sys.exit("Error: wrong objective field file syntax\n%s" %
+                             ",".join(row))
+                weights = row[:]
+                try:
+                    weights[1] = int(weights[1])
+                except ValueError:
+                    sys.exit("Error: wrong objective field file syntax\n%s" %
+                             ",".join(row))
+                objective_weights.append(weights)
+            return objective_weights
+    except IOError:
+        sys.exit("Error: cannot read objective weights %s" % path)
+
+
+def is_shared(resource):
+    """Checks if a resource is shared
+
+    """
+    return resource['object'].get('shared', False)
+
+
+def get_options_list(args, options, prioritary=None):
+    """Creates the list of values that generates the options in `options`
+       from the args object.
+
+    """
+    options_list = []
+    filtered_options = options[:]
+    non_equivalent_options = {'dev': 'dev_mode'}
+    # removes the options that are already set as prioritary
+    for element in prioritary:
+        try:
+            filtered_options.remove(element)
+        except ValueError:
+            pass
+    # adds the options that have not been assigned a value in prioritary
+    for option in filtered_options:
+        try:
+            flag = option[2:].replace("-", "_")
+            if flag in non_equivalent_options.keys():
+                flag = non_equivalent_options[flag]
+            value = getattr(args, flag)
+            if value is not None:
+                if isinstance(value, bool):
+                    if value:
+                        options_list.append(option)
+                elif isinstance(value, list):
+                    if value:
+                        options_list.append("%s=%s" %
+                                            (option, ",".join(value)))
+                else:
+                    if not isinstance(value, basestring):
+                        value = str(value)
+                    options_list.append("%s=%s" % (option, value))
+        except AttributeError:
+            pass
+    return options_list
+
+
+def print_generated_files(path, log_file=None, verbosity=1):
+    """Prints the file structure generated while running bigmler
+
+    """
+    message = (u"\nGenerated files:\n\n" +
+               print_tree(path, u" ") + u"\n")
+    log_message(message, log_file=log_file, console=verbosity)
+
+
+def storage_file_name(directory, resource_id):
+    """Returns the path to the JSON file that contains a stored resource
+       structure
+
+    """
+    return os.path.normpath(
+        os.path.join(directory, resource_id.replace("/", "_")))
+
+
+def get_objective_id(fields, objective):
+    """Checks if the objective given by the user in the --objective flag
+       is in the list of fields. Returns its column number or None otherwise.
+
+    """
+    if objective is None:
+        return None
+    try:
+        objective_id = fields.field_id(objective)
+    except ValueError:
+        return None
+    return objective_id
+
+
+def open_mode(mode):
+    """Python 3 compatible open mode
+
+    """
+    return "%st" % mode if PYTHON3 else "%s" % mode
+
+
+def encode2(value, encoding=FILE_ENCODING):
+    """Conditional encoding only for Python2
+
+    """
+    if isinstance(value, basestring) and not PYTHON3:
+        return value.encode(encoding)
+    return value
+
+
+def decode2(value, encoding=FILE_ENCODING):
+    """Conditional decoding only for Python2
+
+    """
+    if isinstance(value, basestring) and not PYTHON3:
+        return value.decode(encoding)
+    return value
+
+
+def transform_fields_keys(json_attributes, fields):
+    """Transforms the fields structure keys if they are expressed in columns
+       to the corresponding IDs
+
+    """
+    fields_structure = {}
+    if fields is None:
+        return json_attributes
+    if "fields" in json_attributes.keys():
+        old_keys = json_attributes["fields"].keys()
+        for old_key in old_keys:
+            try:
+                if not old_key in fields.fields:
+                    key = fields.field_id(int(old_key))
+                else:
+                    key = old_key
+            except ValueError:
+                key = old_key
+            fields_structure[key] = json_attributes["fields"][old_key]
+        json_attributes["fields"] = fields_structure
+    return json_attributes
